from sentence_transformers.util import pytorch_cos_sim
import torch

#Parent class for any dense model
class DenseRetrieval():
    
    def __init__(self, embedder, **kwargs):
        #embedder is class that provides encode_corpus() and encode_queries()
        self.embedder = embedder
        self.batch_size = 128
        self.corpus_chunk_size = 50000
        self.show_progress_bar = True
        self.results = {}
    
    def hnswlib_search(self, corpus, queries):
        pass
    
    def faiss_search(self, corpus, queries):
        pass
    
    def exact_search(self, corpus, queries, top_k):
        #Create embeddings for all queries using embedder.encode_queries()
        #Runs semantic search against the corpus embeddings
        #Returns a ranked list with the corpus ids
            
        print("Encoding Queries...")
        query_ids = list(queries.keys())
        self.results = {qid: {} for qid in query_ids}
        queries = [queries[qid]["text"] for qid in queries]
        query_embeddings = self.embedder.encode_queries(
            queries, show_progress_bar=self.show_progress_bar, batch_size=self.batch_size)
          
        print("Encoding Corpus...")
        corpus_ids = list(corpus.keys())
        corpus = [corpus[cid] for cid in corpus_ids]
            
        itr = range(0, len(corpus), self.corpus_chunk_size)
                
        for corpus_start_idx in itr:
            corpus_end_idx = min(corpus_start_idx + self.corpus_chunk_size, len(corpus))

            #Encode chunk of corpus    
            sub_corpus_embeddings = self.embedder.encode_corpus(
                corpus[corpus_start_idx:corpus_end_idx],
                show_progress_bar=self.show_progress_bar, 
                batch_size=self.batch_size
                )

            #Compute cosine similarites
            cos_scores = pytorch_cos_sim(query_embeddings, sub_corpus_embeddings)
            cos_scores[torch.isnan(cos_scores)] = -1

            #Get top-k values
            cos_scores_top_k_values, cos_scores_top_k_idx = torch.topk(cos_scores, min(top_k+1, len(cos_scores[0])), dim=1, largest=True, sorted=False)
            cos_scores_top_k_values = cos_scores_top_k_values.cpu().tolist()
            cos_scores_top_k_idx = cos_scores_top_k_idx.cpu().tolist()
            
            for query_itr in range(len(query_embeddings)):
                query_id = query_ids[query_itr]                  
                for sub_corpus_id, score in zip(cos_scores_top_k_idx[query_itr], cos_scores_top_k_values[query_itr]):
                    corpus_id = corpus_ids[corpus_start_idx+sub_corpus_id]
                    if corpus_id != query_id:
                        self.results[query_id][corpus_id] = score
        
            return self.results 



import pytrec_eval

from search import *
from models import *


class EvaluateRetrieval:
    
    def __init__(self, model, model_name, k_values=[1,3,5,10,100,1000]):
        self.k_values = k_values
        self.top_k = max(k_values)
        if model not in ["bm25"]:
            self.retrieval = DenseRetrieval(SentenceBERT(model_name))
    
    def evaluate(self, corpus, queries, qrels):
        
        results = self.retrieval.exact_search(corpus, queries, self.top_k)
        return self.rank(qrels, results, self.k_values)
    
    @staticmethod
    def rank(qrels, results, k_values):
    
        ndcg = {}
        map_ = {}
        recall = {}
        
        for k in k_values:
            ndcg[k] = 0.0
            map_[k] = 0.0
            recall[k] = 0.0
        
        map_string = "map_cut." + ",".join([str(k) for k in k_values])
        ndcg_string = "ndcg_cut." + ",".join([str(k) for k in k_values])
        recall_string = "recall." + ",".join([str(k) for k in k_values])
        evaluator = pytrec_eval.RelevanceEvaluator(qrels, {map_string, ndcg_string, recall_string})
        scores = evaluator.evaluate(results)
        
        for query_id in scores.keys():
            for k in k_values:
                ndcg[k] += scores[query_id]["ndcg_cut_" + str(k)]
                map_[k] += scores[query_id]["map_cut_" + str(k)]
                recall[k] += scores[query_id]["recall_" + str(k)]
        
        for k in k_values:
            ndcg[k] = round(ndcg[k]/len(scores), 5)
            map_[k] = round(map_[k]/len(scores), 5)
            recall[k] = round(recall[k]/len(scores), 5)
        
        return {"ndcg@k": ndcg, "map@k": map_, "recall@k": recall}
    


from elasticsearch import Elasticsearch
from elasticsearch.helpers import streaming_bulk
import logging
import tqdm
import sys

tracer = logging.getLogger('elasticsearch') 
tracer.setLevel(logging.CRITICAL) # supressing INFO messages for elastic-search

class ElasticSearch(object):
    
    def __init__(self, hostname, index_name, text_key, title_key, timeout=5000):
        logging.info("Activating Elastic Search....")
        self.index_name = index_name
        self.text_key = text_key
        self.title_key = title_key
        self.es = Elasticsearch([hostname], timeout=timeout, retry_on_timeout=True, maxsize=24)
        
    
    def create_index(self):
        """Create Elasticsearch Index

        Args:
            dims (int): dimension of dense vector
        """
        logging.info("Creating fresh Elasticsearch-Index named - {}".format(self.index_name))
        
        try:
            mapping = {
                "mappings" : {
                    "properties" : {
                        self.title_key: {"type": "text"},
                        self.text_key: {"type": "text"}
                    }
                }
            }
            self.es.indices.create(index=self.index_name, body=mapping, ignore=[400]) #400: IndexAlreadyExistsException
        except Exception as e:
            logging.error("Unable to create Index in Elastic Search. Reason: {}".format(e))
    
    def delete_index(self):
        """Delete Elasticsearch Index"""
        
        logging.info("Deleting previous Elasticsearch-Index named - {}".format(self.index_name))
        try:
            self.es.indices.delete(index=self.index_name, ignore=[400, 404]) # 404: IndexDoesntExistException
        except Exception as e:
            logging.error("Unable to create Index in Elastic Search. Reason: {}".format(e))
    
    def bulk_add_to_index(self, generate_actions, progress):
        """Bulk indexing to elastic search using generator actions

        Args:
            generate_actions (generator function): generator function must be provided 
            progress (tqdm.tqdm): tqdm progress_bar
        """
        for ok, action in streaming_bulk(
        client=self.es, index=self.index_name, actions=generate_actions,
        ):
            progress.update(1)
        progress.reset()
        progress.close()
    
    def lexical_multisearch(self, texts, top_hits, skip=0, text_present=False):
        """lexical search using text in Elastic Search

        Args:
            id (str): id of the head paragraph
            text (str): text of the head paragraph
            top_hits (int): top k hits to retrieved

        Returns:
            dict: Hit results
        """
        request = []
        
        assert top_hits + skip <= 10000, "Elastic-Search Window too large, Max-Size = 10000"
        
        if top_hits == 10000 and skip == 0 and text_present == False:
            top_hits = 9999
        
        for text in texts:
            req_head = {"index" : self.index_name, "search_type": "dfs_query_then_fetch"}
            req_body = {
                "_source": False, # No need to return source objects
                "query": {
                    "multi_match": { 
                        "query": text, # matching query with both text and title fields
                        "type": "best_fields",
                        "fields": [self.title_key, self.text_key],
                        "tie_breaker": 0.5
                        }
                    },
                "size": top_hits + skip + 1, # The same paragraph will occur in results
                }
            request.extend([req_head, req_body])
        
        res = self.es.msearch(body = request)

        result = []
        for resp in res["responses"]:
            if text_present:
                responses = resp["hits"]["hits"][(1+skip):]
            else:
                responses = resp["hits"]["hits"][skip:-1]
            
            hits = []
            for hit in responses:
                hits.append((hit["_id"], hit['_score']))

            result.append(self.hit_template(es_res=resp, hits=hits))
        return result
    
    
    def generate_actions(self, dictionary, update=False):
        """Iterator function for efficient addition to elastic-search
        Update - https://stackoverflow.com/questions/35182403/bulk-update-with-pythons-elasticsearch

        Args:
            dictionary (dict): dictionary with id and sentence_text/sentence_embedding.
        """
        for _id, value in dictionary.items():
            if not update: 
                doc = {
                    "_id": str(_id),
                    "_op_type": "index",
                    self.text_key: value[self.text_key],
                    self.title_key: value[self.title_key],
                }
            else:
                doc = {
                    "_id": str(_id),
                    "_op_type": "update",
                    "doc": {
                        self.text_key: value[self.text_key],
                        self.title_key: value[self.title_key],
                        }
                }
                
            yield doc
        
    def hit_template(self, es_res, hits):
        result = {
            'meta': {
                'total': es_res['hits']['total']['value'],
                'took': es_res['took'],
                'num_hits': len(hits)
            },
            'hits': hits,
        }
        return result


import torch
from transformers import DPRContextEncoder, DPRContextEncoderTokenizer
from transformers import DPRQuestionEncoder, DPRQuestionEncoderTokenizer

class DPR:
    def __init__(self, model_name_or_path):
        
        # Query tokenizer and model
        self.query_tokenizer = DPRQuestionEncoderTokenizer.from_pretrained('facebook/dpr-question_encoder-single-nq-base')
        self.query_model = DPRQuestionEncoder.from_pretrained('facebook/dpr-question_encoder-single-nq-base')
        
        # Context tokenizer and model
        self.context_tokenizer = DPRContextEncoderTokenizer.from_pretrained('facebook/dpr-ctx_encoder-single-nq-base')
        self.context_model = DPRContextEncoder.from_pretrained('facebook/dpr-ctx_encoder-single-nq-base')
    
    def encode_query(self, sentences, **kwargs):
        output = []
        batch_size = 16
        with torch.no_grad():
            for start_idx in tqdm.trange(0, len(texts), batch_size, desc='que'):
                encoded = self.query_tokenizer(texts[start_idx:start_idx+batch_size], truncation=True, padding=True, return_tensors='pt')
                model_out = self.query_model(encoded['input_ids'].cuda(), attention_mask=encoded['attention_mask'].cuda())
                embeddings_q = model_out.pooler_output
                for emb in embeddings_q:
                    output.append(emb)

        out_tensor = torch.stack(output)
        assert out_tensor.shape[0] == len(sentences)
        return out_tensor
    
        
    def encode_corpus(self, sentences, **kwargs):
        output = []
        batch_size = 8
        with torch.no_grad():
            for start_idx in tqdm.trange(0, len(sentences), batch_size, desc='pas'):
                titles = [row['title'] for row in sentences[start_idx:start_idx+batch_size]]
                bodies = [row['body']  for row in sentences[start_idx:start_idx+batch_size]]
                encoded = self.tokenizer(titles, bodies, truncation='longest_first', padding=True, return_tensors='pt')
                model_out = self.model(encoded['input_ids'].cuda(), attention_mask=encoded['attention_mask'].cuda())
                embeddings_q = model_out.pooler_output.detach()
                for emb in embeddings_q:
                    output.append(emb)

        out_tensor = torch.stack(output)
        assert out_tensor.shape[0] == len(sentences)
        return out_tensor


from setuptools import setup, find_packages

with open("README.md", mode="r", encoding="utf-8") as readme_file:
    readme = readme_file.read()

setup(
    name="beir",
    version="0.0.2",
    author="Nandan Thakur",
    author_email="nandant@gmail.com",
    description="A Heterogeneous Benchmark for Information Retrieval",
    long_description=readme,
    long_description_content_type="text/markdown",
    license="Apache License 2.0",
    url="https://github.com/beir-nlp/beir",
    download_url="https://github.com/beir-nlp/beir/archive/v0.0.1.zip",
    packages=find_packages(),
    python_requires='>=3.6',
    install_requires=[
        'sentence-transformers',
        'pytrec_eval'
    ],
    classifiers=[
        "Development Status :: 4 - Beta",
        "Intended Audience :: Science/Research",
        "License :: OSI Approved :: Apache Software License",
        "Programming Language :: Python :: 3.6",
        "Topic :: Scientific/Engineering :: Artificial Intelligence"
    ],
    keywords="Information Retrieval Transformer Networks BERT PyTorch IR NLP deep learning"
)


from elastic_search import ElasticSearch

class BM25:
    def __init__(self, hostname, index_name, text_key, title_key, timeout=100):
        print(hostname)
        self.es = ElasticSearch(hostname=hostname, index_name=index_name, 
                                text_key=text_key, title_key=title_key)
        self.initialise()
    
    def initialise(self):
        self.es.delete_index()
        self.es.create_index()
    
    def bm25_index(self, corpus, titles, text_key, title_key):
        progress = tqdm.tqdm(unit="docs", total=len(corpus))
        # dictionary structure = {_id: {title_key: title, text_key: text}}
        dictionary = {idx: {
            title_key: titles.get(idx, None), 
            text_key: corpus.get(idx, None)
            } for idx in list(corpus.keys())
        }
        self.es.bulk_add_to_index(
                generate_actions=self.es.generate_actions(
                dictionary=dictionary, update=False),
                progress=progress
                )


    def bm25_rank(self, queries, judgements, top_k, batch_size):
        """[summary]

        Args:
            queries ([type]): [description]
            top_k ([type]): [description]
            batch_size ([type]): [description]
        """
        rank_results = {}
        
        _type = next(iter(list(judgements.values())[0]))
        generator = chunks(list(queries.keys()), batch_size)
        batches = int(len(queries)/batch_size)
        total = batches if len(queries) % batch_size == 0 else batches + 1 
        
        for query_id_chunks in tqdm.tqdm(generator, total=total):
            texts = [queries[query_id] for query_id in query_id_chunks]
            results = self.es.lexical_multisearch(
                texts=texts, top_hits=top_k + 1) 
            # add 1 extra just incase if query within document

            for (query_id, hit) in zip(query_id_chunks, results):
                scores = {}
                for corpus_id, score in hit['hits']:
                    corpus_id = type(_type)(corpus_id)
                    # making sure query doesnt return in results
                    if corpus_id != query_id:
                        scores[corpus_id] = score
                    
                rank_results[query_id] = scores
                
        return rank_results


from sentence_transformers import SentenceTransformer
import numpy as np

class SentenceBERT:
    def __init__(self, model_path):
        self.model = SentenceTransformer(model_path)
    
    def encode_queries(self, queries, **kwargs):
        return np.asarray(self.model.encode(queries, **kwargs))
    
    def encode_corpus(self, corpus, **kwargs):
        sentences = [(doc["title"] + "  " + doc["text"]).strip() for doc in corpus]   
        return np.asarray(self.model.encode(sentences, **kwargs))